import logging
from typing import Optional

from django.conf import settings
from django.db.models import Q
from django_rq import job
from rq import get_current_job

from arkindex.dataimport.models import DataImport, DataImportElement, WorkerActivity, WorkerRun
from arkindex.documents.managers import ElementQuerySet
from arkindex.documents.models import (
    Classification,
    Corpus,
    Element,
    ElementPath,
    EntityLink,
    MetaData,
    Selection,
    Transcription,
    TranscriptionEntity,
)

logger = logging.getLogger(__name__)


@job('high', timeout=settings.RQ_TIMEOUTS['corpus_delete'])
def corpus_delete(corpus_id: str) -> None:
    # Note that this can be None when the task is run outside of a RQ worker (e.g. unit test)
    rq_job = get_current_job()
    corpus = Corpus.objects.get(id=corpus_id)
    logger.info(f'Deleting {corpus!r}')

    # When a corpus has a top level type, deleting all the ElementTypes will cause an IntegrityError.
    if corpus.top_level_type_id is not None:
        corpus.top_level_type_id = None
        corpus.save(update_fields=['top_level_type_id'])

    # Delete all related objects, bypassing RESTRICTs deletion rules
    # and Django's way of loading everything into memory before deleting.
    querysets = [
        corpus.allowed_metadatas.all(),
        corpus.worker_version_cache.all(),
        WorkerRun.objects.filter(dataimport__corpus_id=corpus_id),
        # Using two separate queries is a lot faster than using a OR for large corpora
        DataImportElement.objects.filter(dataimport__corpus_id=corpus_id),
        DataImportElement.objects.filter(element__corpus_id=corpus_id),
        # DataImport-DataFile M2M with implicit model
        DataImport.files.through.objects.filter(dataimport__corpus_id=corpus_id),
        DataImport.files.through.objects.filter(datafile__corpus_id=corpus_id),
        # Worker activities
        WorkerActivity.objects.filter(Q(element__corpus_id=corpus_id) | Q(process__corpus_id=corpus_id)),
        corpus.imports.all(),
        corpus.files.all(),
        MetaData.objects.filter(element__corpus_id=corpus_id),
        EntityLink.objects.filter(role__corpus_id=corpus_id),
        corpus.roles.all(),
        TranscriptionEntity.objects.filter(entity__corpus_id=corpus_id),
        TranscriptionEntity.objects.filter(transcription__element__corpus_id=corpus_id),
        corpus.entities.all(),
        Classification.objects.filter(element__corpus_id=corpus_id),
        corpus.ml_classes.all(),
        Transcription.objects.filter(element__corpus_id=corpus_id),
        ElementPath.objects.filter(element__corpus_id=corpus_id),
        Selection.objects.filter(element__corpus_id=corpus_id),
        corpus.elements.all(),
        corpus.types.all(),
        corpus.memberships.all(),
        corpus.exports.all(),
        Corpus.objects.filter(id=corpus_id),
    ]

    for i, queryset in enumerate(querysets):
        if rq_job:
            rq_job.set_progress(i / len(querysets))
            rq_job.set_description(f'Deleting {queryset.model.__name__} on corpus {corpus.name}')
        deleted_count = queryset._raw_delete(using='default')
        logger.info(f'Deleted {deleted_count} {queryset.model.__name__}')

    logger.info(f'Deleted corpus {corpus_id}')


@job('high', timeout=settings.RQ_TIMEOUTS['element_trash'])
def element_trash(queryset: ElementQuerySet, delete_children: bool) -> None:
    queryset.trash(delete_children=delete_children)


@job('high', timeout=settings.RQ_TIMEOUTS['worker_results_delete'])
def worker_results_delete(corpus_id: str, version_id: Optional[str], parent_id: Optional[str]) -> None:
    """
    Recursively delete all Worker Results produced by any WorkerVersion or a specific one
    on a whole corpus or under a specified parent element (parent element included).
    """
    elements = Element.objects.filter(corpus_id=corpus_id)
    classifications = Classification.objects.filter(element__corpus_id=corpus_id)
    transcriptions = Transcription.objects.filter(element__corpus_id=corpus_id)
    # TranscriptionEntities that will be removed due to the removal of transcriptions
    transcription_entities = TranscriptionEntity.objects.filter(transcription__element__corpus_id=corpus_id)
    # TranscriptionEntities from this worker
    worker_transcription_entities = TranscriptionEntity.objects.filter(
        Q(transcription__element__corpus_id=corpus_id)
        | Q(entity__corpus_id=corpus_id)
    )
    metadata = MetaData.objects.filter(element__corpus_id=corpus_id)
    worker_activities = WorkerActivity.objects.filter(element__corpus_id=corpus_id)

    # When a version ID is defined, filter by the exact version ID
    if version_id:
        elements = elements.filter(worker_version_id=version_id)
        classifications = classifications.filter(worker_version_id=version_id)
        transcriptions = transcriptions.filter(worker_version_id=version_id)
        transcription_entities = transcription_entities.filter(transcription__worker_version_id=version_id)
        worker_transcription_entities = worker_transcription_entities.filter(worker_version_id=version_id)
        metadata = metadata.filter(worker_version_id=version_id)
        worker_activities = worker_activities.filter(worker_version_id=version_id)
    # Otherwise, select everything that has any worker version ID.
    # We use worker_version_id != None and not worker_version_id__isnull=False,
    # because isnull would cause an unnecessary LEFT JOIN query.
    else:
        elements = elements.exclude(worker_version_id=None)
        classifications = classifications.exclude(worker_version_id=None)
        transcriptions = transcriptions.exclude(worker_version_id=None)
        transcription_entities = transcription_entities.exclude(transcription__worker_version_id=None)
        worker_transcription_entities = worker_transcription_entities.exclude(worker_version_id=None)
        metadata = metadata.exclude(worker_version_id=None)
        worker_activities = worker_activities.exclude(worker_version_id=None)

    # When a parent ID is defined, filter using element paths to match the parent and all its descendants
    if parent_id:
        elements = elements.filter(Q(id=parent_id) | Q(paths__path__overlap=[parent_id]))
        classifications = classifications.filter(Q(element_id=parent_id) | Q(element__paths__path__overlap=[parent_id]))
        transcriptions = transcriptions.filter(Q(element_id=parent_id) | Q(element__paths__path__overlap=[parent_id]))
        # The same filter is applied to both TranscriptionEntity querysets
        transcription_entities_filter = Q(transcription__element_id=parent_id) | Q(transcription__element__paths__path__overlap=[parent_id])
        transcription_entities = transcription_entities.filter(transcription_entities_filter)
        worker_transcription_entities = worker_transcription_entities.filter(transcription_entities_filter)
        metadata = metadata.filter(Q(element_id=parent_id) | Q(element__paths__path__overlap=[parent_id]))
        worker_activities = worker_activities.filter(Q(element_id=parent_id) | Q(element__paths__path__overlap=[parent_id]))

    elements.trash()
    classifications.delete()
    # Delete TranscriptionEntities before transcriptions so that we can delete transcriptions using a single DELETE query
    transcription_entities.delete()
    worker_transcription_entities.delete()
    transcriptions._raw_delete(using='default')
    metadata.delete()
    worker_activities.delete()


@job('high', timeout=settings.RQ_TIMEOUTS['move_element'])
def move_element(source: Element, destination: Element) -> None:
    parents = Element.objects.filter(id__in=source.paths.values('path__last')).only('id')
    for parent in parents:
        parent.remove_child(source)
    source.add_parent(destination)
